use std::{error::Error, fmt::Display, fs::read_to_string};

/// Basic error type
#[derive(Debug)]
enum ParsingError {
    EmptyInstructions,
    ParserFailure,
}

impl Display for ParsingError {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            ParsingError::EmptyInstructions => write!(f, "No instructions present to process"),
            ParsingError::ParserFailure => write!(f, "Failed to parse segment"),
        }
    }
}

impl Error for ParsingError {}

/// Main execution function to analyze memory for uncorrupted mul commands
pub fn analyze_memory(file_path: &str) -> Result<(), Box<dyn Error>> {
    let instructions = read_to_string(file_path)?;
    let part1_sol = process_mul_instructions(&instructions)?;
    println!("Solution to part1 is \n{part1_sol}");

    let part2_sol = process_do_dont_instructions(&instructions)?;
    println!("Solution to part2 is \n{part2_sol}");

    Ok(())
}

///Process mul instructions
pub fn process_mul_instructions(instructions: &str) -> Result<i32, Box<dyn Error>> {
    let segments = segment_instructions(instructions, 'm')?;
    Ok(segments
        .iter()
        .map(|segment| calculate_segment_value(segment).unwrap_or(0))
        .sum::<i32>())
}
/// Execution code for solving part 2 of the problem
pub fn process_do_dont_instructions(instructions: &str) -> Result<i32, Box<dyn Error>> {
    let segments = segment_instructions(instructions, 'd')?;
    let mut last_was_do = true;
    let mut sum = 0;
    for segment in segments.iter() {
        if segment.starts_with("don't()") {
            last_was_do = false;
        } else if segment.starts_with("do()") {
            last_was_do = true;
            sum += process_mul_instructions(segment).unwrap_or(0);
        } else if last_was_do {
            sum += process_mul_instructions(segment).unwrap_or(0);
        }
    }
    Ok(sum)
}

/// Break up the total memory into separated strings for further processing
pub fn segment_instructions(
    instructions: &str,
    filter_char: char,
) -> Result<Vec<String>, Box<dyn Error>> {
    let char_indices: Vec<usize> = instructions
        .chars()
        .enumerate()
        .filter_map(|(i, char)| if char == filter_char { Some(i) } else { None })
        .collect();

    // Include the first chunk by default
    let mut chunks: Vec<String> = [0]
        .iter()
        .chain(char_indices.iter())
        .zip(char_indices.iter())
        .map(|(idx1, idx2)| {
            String::from_utf8_lossy(&instructions.as_bytes()[*idx1..*idx2]).to_string()
        })
        .collect();

    chunks.push(
        String::from_utf8_lossy(
            &instructions.as_bytes()
                [*char_indices.last().ok_or(ParsingError::EmptyInstructions)?..],
        )
        .to_string(),
    );
    Ok(chunks)
}

/// Analyze each segment to see if it's valid, if it is valid then return the value of the
/// multiplication. If it isn't valid we return an error.
/// We could use Regex here but I wanted to try a more manual approach to keep with the spirit of
/// minimizing external libraries
pub fn calculate_segment_value(segment: &str) -> Result<i32, Box<dyn Error>> {
    // Verify string starts correctly
    if !segment.starts_with("mul(") {
        Err(Box::new(ParsingError::ParserFailure))
    } else {
        let mut segment = segment.to_owned();
        segment.drain(..4);
        let comma_split: Vec<&str> = segment.split(",").collect();
        let parens_split: Vec<&str> = comma_split
            .get(1)
            .ok_or(ParsingError::ParserFailure)?
            .split(")")
            .collect();

        if parens_split.len() < 2 {
            return Err(Box::new(ParsingError::ParserFailure));
        };

        if let (Ok(left), Ok(right)) = (
            comma_split[0].parse::<i32>(),
            parens_split[0].parse::<i32>(),
        ) {
            Ok(left * right)
        } else {
            Err(Box::new(ParsingError::ParserFailure))
        }
    }
}

#[cfg(test)]
mod test {
    use super::*;

    #[test]
    pub fn test_segmentation() {
        let test_string = "lkjqermulemtymart";
        let segments = segment_instructions(test_string, 'm').unwrap();
        assert_eq!(segments[0], "lkjqer");
        assert_eq!(segments[1], "mule");
        assert_eq!(segments[2], "mty");
        assert_eq!(segments[3], "mart");
    }

    #[test]
    pub fn test_segment_value() {
        assert_eq!(
            calculate_segment_value("mul(10,14)aeqejrqlkd9()").unwrap(),
            140
        );
        assert_eq!(
            calculate_segment_value("mul(1a0,14)aeqejrqlkd9()").unwrap_or(0),
            0
        );
        assert_eq!(
            calculate_segment_value("mul(10,14aeqejrqlkd9()").unwrap_or(0),
            0
        );

        assert_eq!(
            calculate_segment_value("mul(910,54,don't(),#^+{").unwrap_or(0),
            0
        );
    }

    #[test]
    pub fn test_do_dont() {
        let instructions =
            "mul(1,2)don't()mul(2,3)ekjkldomul(2,3)do()mul(211,adf2),))),)do()mul(211,12)";
        assert_eq!(process_do_dont_instructions(instructions).unwrap(), 2534);
    }
}
