use std::{
    collections::HashSet,
    error::Error,
    fmt::Display,
    fs::File,
    io::{BufRead, BufReader},
    rc::Rc,
};

/// Basic error type
#[derive(Debug)]
enum ParsingError {
    ParserFailure,
    IndexOutOfBounds,
    InconsistentGrid,
}

/// Enum dictating the direction of the guard
#[derive(Clone)]
pub enum GuardDirection {
    Up,
    Right,
    Down,
    Left,
}

impl GuardDirection {
    // Cycle through directions when guard hits an obstacle
    pub fn cycle(&self) -> Self {
        match self {
            GuardDirection::Up => GuardDirection::Right,
            GuardDirection::Right => GuardDirection::Down,
            GuardDirection::Down => GuardDirection::Left,
            GuardDirection::Left => GuardDirection::Up,
        }
    }
}

impl Display for ParsingError {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            ParsingError::ParserFailure => write!(f, "Failed to parse segment"),
            ParsingError::IndexOutOfBounds => write!(f, "Index out of bounds for CharGrid"),
            ParsingError::InconsistentGrid => write!(f, "File is not a consistent char grid"),
        }
    }
}

type MyResult<T> = Result<T, Box<dyn Error>>;

impl Error for ParsingError {}

/// Struct that flattens out the input grid
#[derive(Clone)]
pub struct CharGrid {
    pub chars: Rc<Vec<char>>,
    pub m: usize,
    pub n: usize,
    pub guard_position: usize,
    pub visited_positions: HashSet<usize>,
    pub guard_direction: GuardDirection,
}

impl CharGrid {
    /// Take a file and split it into a CharGrid. Since input has only ASCII characters this should
    /// be fine
    pub fn new(input_file: &str) -> MyResult<CharGrid> {
        let file = File::open(input_file)?;
        let reader = BufReader::new(file);

        let mut m = 0;
        let mut n: Option<usize> = None;
        let chars: Vec<char> = reader
            .lines()
            .map(|line| -> MyResult<Vec<char>> {
                let parsed_line = Vec::from_iter(line?.chars());
                m += 1;
                if n.is_none() {
                    n = Some(parsed_line.len());
                } else if n.unwrap() != parsed_line.len() {
                    return Err(Box::new(ParsingError::InconsistentGrid));
                }
                Ok(parsed_line)
            })
            .collect::<MyResult<Vec<Vec<char>>>>()?
            .into_iter()
            .flatten()
            .collect();

        let guard_position = chars.iter().position(|&x| x == '^').unwrap();

        // Safe to unwrap n because it is validated if we get to this point
        if chars.len() != m * n.unwrap() {
            Err(Box::new(ParsingError::ParserFailure))
        } else {
            Ok(CharGrid {
                chars: Rc::new(chars),
                m,
                n: n.unwrap(),
                guard_position,
                guard_direction: GuardDirection::Up,
                visited_positions: HashSet::new(),
            })
        }
    }

    /// Get an index within the grid
    pub fn get(&self, i: i32, j: i32) -> MyResult<char> {
        let idx = self.two2one_index(i, j)?;
        Ok(self.chars[idx])
    }

    // Convert the one dimensional index to two dimensional index
    pub fn one2two_index(&self, idx: usize) -> (usize, usize) {
        let i = idx / self.n;
        let j = idx % self.n;
        (i, j)
    }

    // Convert the two dimensional index to one dimensional index
    pub fn two2one_index(&self, i: i32, j: i32) -> MyResult<usize> {
        if i < 0 || j < 0 {
            Err(Box::new(ParsingError::IndexOutOfBounds))
        } else if i as usize >= self.m || j as usize >= self.n {
            Err(Box::new(ParsingError::IndexOutOfBounds))
        } else {
            Ok((i as usize) * self.n + (j as usize))
        }
    }
}

struct MoveGuardResult {
    new_grid: CharGrid,
    guard_exited: bool,
}

// Moves the guard on their current path until an obstacle or they leave the map
fn move_guard_on_path(grid: &CharGrid) -> MoveGuardResult {
    let (guard_i, guard_j) = grid.one2two_index(grid.guard_position);
    let to_visit: Vec<(i32, i32)> = match grid.guard_direction {
        GuardDirection::Up => (-1..guard_i as i32)
            .rev()
            .map(|i| (i, guard_j as i32))
            .collect(),
        GuardDirection::Right => (guard_j as i32..grid.n as i32 + 1)
            .map(|j| (guard_i as i32, j))
            .collect(),
        GuardDirection::Down => (guard_i as i32..grid.m as i32 + 1)
            .map(|i| (i, guard_j as i32))
            .collect(),
        GuardDirection::Left => (-1..guard_j as i32)
            .rev()
            .map(|j| (guard_i as i32, j))
            .collect(),
    };
    let mut new_grid = grid.clone();

    println!(
        "{:?} {}",
        new_grid.one2two_index(new_grid.guard_position),
        new_grid.guard_direction.clone() as usize
    );
    let mut last_pos = grid.guard_position;
    for (i, j) in to_visit.into_iter() {
        let c = match grid.get(i, j) {
            Ok(val) => val,
            // Guard left the grid
            Err(_) => {
                println!("Failure");
                return MoveGuardResult {
                    new_grid,
                    guard_exited: true,
                };
            }
        };

        if c == '#' {
            new_grid.guard_position = last_pos;
            new_grid.guard_direction = new_grid.guard_direction.cycle();

            return MoveGuardResult {
                new_grid,
                guard_exited: false,
            };
        } else {
            new_grid
                .visited_positions
                .insert(grid.two2one_index(i, j).unwrap());

            last_pos = grid.two2one_index(i, j).unwrap();
        }
    }

    MoveGuardResult {
        new_grid,
        guard_exited: false,
    }
}

pub fn calc_unique_spots(grid: &CharGrid) -> usize {
    let mut result = move_guard_on_path(grid);
    while !result.guard_exited {
        result = move_guard_on_path(&result.new_grid);
    }

    // +1 to account for intial guard position
    result.new_grid.visited_positions.len() + 1
}

#[cfg(test)]
mod test {
    use super::*;

    #[test]
    fn test_char_grid() {
        let grid = CharGrid::new("test_inputs/test_input.txt").unwrap();
    }

    #[test]
    fn test_guard_calc() {
        let grid = CharGrid::new("test_inputs/test_input.txt").unwrap();

        assert_eq!(calc_unique_spots(&grid), 41);
    }
}
