use std::{
    collections::{HashMap, HashSet},
    error::Error,
    fmt::Display,
    fs::File,
    io::{BufRead, BufReader},
    rc::Rc,
};

use itertools::Itertools;

/// Basic error type
#[derive(Debug)]
enum ParsingError {
    ParserFailure,
    IndexOutOfBounds,
    InconsistentGrid,
}

impl Display for ParsingError {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            ParsingError::ParserFailure => write!(f, "Failed to parse segment"),
            ParsingError::IndexOutOfBounds => write!(f, "Index out of bounds for CharGrid"),
            ParsingError::InconsistentGrid => write!(f, "File is not a consistent char grid"),
        }
    }
}

type MyResult<T> = Result<T, Box<dyn Error>>;

impl Error for ParsingError {}

/// Struct that flattens out the input grid
#[derive(Clone)]
pub struct CharGrid {
    pub chars: Vec<char>,
    pub m: usize,
    pub n: usize,
    pub same_freq_positions: Vec<Vec<usize>>,
}

impl CharGrid {
    /// Take a file and split it into a CharGrid. Since input has only ASCII characters this should
    /// be fine
    pub fn new(input_file: &str) -> MyResult<CharGrid> {
        let file = File::open(input_file)?;
        let reader = BufReader::new(file);

        let mut m = 0;
        let mut n: Option<usize> = None;
        let chars: Vec<char> = reader
            .lines()
            .map(|line| -> MyResult<Vec<char>> {
                let parsed_line = Vec::from_iter(line?.chars());
                m += 1;
                if n.is_none() {
                    n = Some(parsed_line.len());
                } else if n.unwrap() != parsed_line.len() {
                    return Err(Box::new(ParsingError::InconsistentGrid));
                }
                Ok(parsed_line)
            })
            .collect::<MyResult<Vec<Vec<char>>>>()?
            .into_iter()
            .flatten()
            .collect();

        let n = n.unwrap();
        // Now loop and get all the unique indices
        let mut char_indices: HashMap<char, usize> = HashMap::new();
        let mut curr_idx = 0;
        let mut same_freq_positions: Vec<Vec<usize>> = Vec::new();

        for (i, c) in chars.iter().enumerate() {
            if *c != '.' {
                let idx = match char_indices.get(c) {
                    Some(idx) => idx,
                    None => {
                        char_indices.insert(*c, curr_idx);
                        curr_idx += 1;
                        same_freq_positions.push(vec![]);
                        &(curr_idx - 1)
                    }
                };

                same_freq_positions[*idx].push(i);
            }
        }
        Ok(CharGrid {
            chars,
            m,
            n,
            same_freq_positions,
        })
    }

    /// Get an index within the grid
    pub fn get(&self, i: i32, j: i32) -> MyResult<char> {
        let idx = self.two2one_index(i, j)?;
        Ok(self.chars[idx])
    }

    // Convert the one dimensional index to two dimensional index
    pub fn one2two_index(&self, idx: usize) -> (usize, usize) {
        let i = idx / self.n;
        let j = idx % self.n;
        (i, j)
    }

    // Convert the two dimensional index to one dimensional index
    pub fn two2one_index(&self, i: i32, j: i32) -> MyResult<usize> {
        if i < 0 || j < 0 {
            Err(Box::new(ParsingError::IndexOutOfBounds))
        } else if i as usize >= self.m || j as usize >= self.n {
            Err(Box::new(ParsingError::IndexOutOfBounds))
        } else {
            Ok((i as usize) * self.n + (j as usize))
        }
    }
}

/// Get the antinodes for a single frequency assuming only resonant frequencies
pub fn get_antinodes_single_freq_part1(grid: &CharGrid, positions: &[usize]) -> HashSet<usize> {
    let grid_iter = positions.iter();
    let mut unique_position = HashSet::new();
    for (node1, node2) in grid_iter.clone().cartesian_product(grid_iter) {
        if node1 != node2 {
            let (x1, y1) = grid.one2two_index(*node1);
            let (x2, y2) = grid.one2two_index(*node2);

            let x_antinode = (2 * x1).checked_sub(x2);
            let y_antinode = (2 * y1).checked_sub(y2);

            if let (Some(x_antinode), Some(y_antinode)) = (x_antinode, y_antinode) {
                if let Ok(idx) = grid.two2one_index(x_antinode as i32, y_antinode as i32) {
                    unique_position.insert(idx);
                }
            }
        }
    }

    unique_position
}

/// Get the antinodes for a single frequency assuming resonant harmonics
/// NOTE: We shouldn't need two separate functions since part1 is a subset of part2 but running out
/// of time again so leaving it as a separate function for now
pub fn get_antinodes_single_freq_part2(grid: &CharGrid, positions: &[usize]) -> HashSet<usize> {
    let grid_iter = positions.iter();
    let mut unique_position = HashSet::new();
    for (node1, node2) in grid_iter.clone().cartesian_product(grid_iter) {
        if node1 != node2 {
            let mut step = 1;
            loop {
                let (x1, y1) = grid.one2two_index(*node1);
                let (x2, y2) = grid.one2two_index(*node2);

                let xstep = x1 as i32 - x2 as i32;
                let ystep = y1 as i32 - y2 as i32;

                let x_antinode: MyResult<usize> = (x1 as i32 - (step * xstep))
                    .try_into()
                    .map_err(|err| -> Box<dyn Error> { Box::new(err) });

                let y_antinode: MyResult<usize> = (y1 as i32 - (step * ystep))
                    .try_into()
                    .map_err(|err| -> Box<dyn Error> { Box::new(err) });

                if let (Ok(x_antinode), Ok(y_antinode)) = (x_antinode, y_antinode) {
                    if let Ok(idx) = grid.two2one_index(x_antinode as i32, y_antinode as i32) {
                        unique_position.insert(idx);
                    } else {
                        break;
                    }
                } else {
                    break;
                }
                step += 1;
            }
        }
    }

    unique_position
}
pub enum Mode {
    Part1,
    Part2,
}
/// Get antinodes for all frequencies
pub fn get_antinode_count(grid: &CharGrid, mode: &Mode) -> usize {
    let mut unique_positions: HashSet<usize> = HashSet::new();
    for positions in grid.same_freq_positions.iter() {
        match mode {
            Mode::Part1 => {
                unique_positions.extend(get_antinodes_single_freq_part1(grid, positions))
            }
            Mode::Part2 => {
                unique_positions.extend(get_antinodes_single_freq_part2(grid, positions))
            }
        }
    }
    unique_positions.len()
}

#[cfg(test)]
mod test {
    use super::*;
    #[test]
    fn test_antinode_calc() {
        let grid = CharGrid::new("test_inputs/test_input.txt").unwrap();
        assert_eq!(get_antinode_count(&grid, &Mode::Part1), 14);
        assert_eq!(get_antinode_count(&grid, &Mode::Part2), 34);
    }
    #[test]
    fn simple_part2_test() {
        let grid = CharGrid::new("./test_inputs/simple_part2.txt").unwrap();
        assert_eq!(get_antinode_count(&grid, &Mode::Part2), 9);
    }
}
